# Imports from python.  # NOQA
from datetime import datetime


# Imports from Django.
from django.db import connections
from django.db.utils import ConnectionDoesNotExist
from django.utils import timezone


# Imports from other dependencies.
from editorial_staff.models import Hub
from editorial_staff.models import Vertical
from psycopg2.extras import DateTimeTZRange


# Imports from budget.
from budget.models import Item
from budget.models import Package


tz = timezone.get_default_timezone()

# Figure out which database we should use to check query counts
try:
    connections['budget'].cursor()
    separate_db = True
    budget_db = 'budget'
except ConnectionDoesNotExist:
    separate_db = False
    budget_db = 'default'


def staff_factory(hub='hub'):
    vertical, created = Vertical.objects.get_or_create(
        slug='vertical',
        name='vertical'
    )

    Hub.objects.get_or_create(
        slug=hub,
        name=hub,
        vertical=vertical
    )


def package_factory(publish_date_lower=tz.localize(datetime(2015, 5, 1)),
                    publish_date_upper=tz.localize(datetime(2015, 5, 2)),
                    print_run_date=None, slug_key='overall-slug'):
    staff_factory()

    return Package.objects.create(
        hub='hub',
        published_url='http://dallasnews.com/story/',
        notes='Notes field',
        publish_date=DateTimeTZRange(lower=publish_date_lower,
                                     upper=publish_date_upper),
        # print_run_date=print_run_date,
        slug_key=slug_key
    )


def item_factory(primary_for=None, additional_for=None, item_type='text',
                 author='a@d.com', editor='e@d.com', slug_key='sluggy-slug',
                 budget_line='Budget line'):
    return Item.objects.create(
        slug_key=slug_key,
        type=item_type,
        editors=[{'email': editor}],
        authors=[{'email': author}],
        budget_line=budget_line,
        primary_for_package=primary_for,
        additional_for_package=additional_for
    )


def get_timestamp_for_date(raw_iso_date):
    return tz.localize(
        datetime.strptime(
            raw_iso_date,
            '%Y-%m-%d'
        )
    ).astimezone(timezone.utc).strftime('%Y-%m-%dT%H:%M:%SZ')



# Imports from Django.  # NOQA
from django.apps import AppConfig


class BudgetConfig(AppConfig):
    name = 'budget'
    verbose_name = 'budget'

    def ready(self):
        from budget import signals  # NOQA



# Imports from Django.  # NOQA
from django.apps import apps
from django.contrib.auth import get_user_model
from django.db import models


# Imports from budget.
# from budget.utils import alphacode
from budget.utils import slug_date_to_range


class ChangeQuerySet(models.QuerySet):
    # Any supported pretch fields should go here, exactly as they'd go into
    # a prefetch_related() call
    CUSTOM_PREFETCH_FIELDS = ('by_user', 'item',)

    def _prefetch_related_objects(self):
        """Filter out custom prefetch fields and read them separately.

        Override the queryset's built-in _prefetch_related_objects
        method to filter out - but then read - the custom prefetch
        fields we support.
        """
        active_custom_prefetches = []

        cleaned_related_lookups = list(self._prefetch_related_lookups)

        for custom_prefetch in self.CUSTOM_PREFETCH_FIELDS:
            if custom_prefetch in self._prefetch_related_lookups:
                active_custom_prefetches.append(custom_prefetch)
                cleaned_related_lookups.remove(custom_prefetch)

        self._prefetch_related_lookups = cleaned_related_lookups
        super(ChangeQuerySet, self)._prefetch_related_objects()

        self._prefetch_related_lookups = (
            self._prefetch_related_lookups + active_custom_prefetches)

    def _prefetch_by_user(self):
        """Prefetch the by_user field.

        Handle this using the PKs in the by model attribute. This needs
        to happen here because the users are in a different table, so we
        can't store a true ForeignKey to them at the model level.

        Instead, we grab them all in a single query and add them onto
        the model's _user attribute, where they're accessed by the
        by_user model method.
        """
        if 'by_user' not in self._prefetch_related_lookups:
            return

        # Get all models that are missing a cached _user property
        missing_users = list(
            [x for x in self._result_cache if not hasattr(x, '_user')]
        )

        if not missing_users:
            return

        # Get all unique user IDs from the models missing _user attributes
        user_ids = [x.by for x in missing_users]
        unique_user_ids = list(set(user_ids))

        # Then get all matching user models
        UserModel = get_user_model()
        users = list(
            UserModel.objects.filter(id__in=unique_user_ids).iterator()
        )

        # Now attach those user models to the results models in our results
        # cache
        def _add_user(change):
            change._user = next(u for u in users if u.id == change.by)
            return change

        self._result_cache = list(map(_add_user, self._result_cache))

    def _prefetch_item(self):
        """Prefetch related budget item from change model.

        Works the same as _prefetch_by_user, but with a slight
        modification: _prefetch_by_user can count on working with a single
        model type. _prefetch_by_item has to query multiple model types.
        """
        if 'item' not in self._prefetch_related_lookups:
            return

        # Get all items that are missing a cached _item property
        missing = list(
            [x for x in self._result_cache if not hasattr(x, '_item')]
        )

        if not missing:
            return

        # Get all unique item IDs and model names for missing _items
        item_ids = list(
            set([(x.item_content_type, x.item_id) for x in missing])
        )
        item_types = list(set([x.item_content_type for x in missing]))

        # Now go through each of the unique model names referenced and load
        # all items of that type needed by this queryset
        prefetched_instances = []

        for model in item_types:
            # Get the model by name
            Model = apps.get_model(app_label='budget', model_name=model)
            # Then get all unique model IDs for matching models
            matches = list([x for x in item_ids if x[0] == model])
            ids_for_model = list([x[1] for x in matches])
            # And run a query to get them all at once
            items = list(Model.objects.filter(id__in=ids_for_model).iterator())

            prefetched_instances += items

        # Tack the results onto the result cache
        for cached in self._result_cache:
            for prefetched in prefetched_instances:
                if (prefetched._meta.model_name == cached.item_content_type and
                        prefetched.id == cached.item_id):
                    setattr(cached, '_item', prefetched)
                    break

    def _fetch_all(self):
        """Fetch all the related fields.

        This is the queryset method that actually evaluates the query,
        so it's the best spot to intercept the result cache before it's
        returned and run our custom prefetch logic.
        """
        super(ChangeQuerySet, self)._fetch_all()

        if not self._result_cache:
            return

        if not isinstance(self._result_cache[0], self.model):
            return

        self._prefetch_by_user()
        self._prefetch_item()


class PackageQuerySet(models.QuerySet):
    """TK."""
    def get_by_slug(self, raw_slug):
        """TK."""
        slug_parts = raw_slug.split('.')

        publish_date_range = slug_date_to_range(slug_parts[2])

        if (publish_date_range.upper - publish_date_range.lower).days == 1:
            return self.get(
                hub=slug_parts[0],
                slug_key=slug_parts[1],
                publish_date__contained_by=publish_date_range
            )

        else:
            Package = apps.get_model(app_label='budget', model_name='package')
            try:
                # Query for month matches first.
                # These will exactly match the given range
                return self.get(
                    hub=slug_parts[0],
                    slug_key=slug_parts[1],
                    publish_date__contained_by=publish_date_range
                )
            except Package.DoesNotExist:
                # If the previous condition failed (and assuming there _is_ an
                # item that matches this date slug), 2 things are true:
                # - This item has a date-granularity of 1 week.
                # - The week to which this item is set spans two months.

                # Instead of looking for an exact match, just find any packages
                # whose dates intersect the current month.
                return self.get(
                    hub=slug_parts[0],
                    slug_key=slug_parts[1],
                    publish_date__contained_by=publish_date_range
                )



# Imports from python.  # NOQA
import os


# Imports from other dependencies.
import dj_database_url


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'fake-key'

INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.postgres',

    'rest_framework',
    'rest_framework.authtoken',
    'django_filters',

    'editorial_staff',
    'budget',
]

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
    },
]

TIME_ZONE = 'America/Chicago'

USE_TZ = True

DATABASES = {}

if 'DATABASE_URL' in os.environ:
    DATABASES['default'] = dj_database_url.config()
else:
    print('ERROR: Please specify a "DATABASE_URL" value in your .env file.')


REST_FRAMEWORK = {
    'DEFAULT_PERMISSION_CLASSES': ('rest_framework.permissions.AllowAny',),
    'DEFAULT_FILTER_BACKENDS': ('rest_framework.filters.DjangoFilterBackend',),
    'PAGE_SIZE': 100
}



# Imports from python.  # NOQA
import json


# Imports from Django.
from django.conf import settings
from django.contrib.staticfiles.templatetags.staticfiles import static
from django.http import Http404
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils import timezone
from django.views.generic import TemplateView
from django.views.generic import View


# Imports from other dependencies.
from djangorestframework_camel_case.parser import CamelCaseJSONParser
from djangorestframework_camel_case.render import CamelCaseJSONRenderer
from rest_framework.authentication import SessionAuthentication
from rest_framework.authentication import TokenAuthentication
from rest_framework.permissions import IsAuthenticated
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework import viewsets
from six.moves.urllib.parse import urlparse
from six.moves.urllib.parse import urlunparse


# Imports from budget.
from budget.filters import ContentPlacementFilter
from budget.filters import HeadlineViewFilter
from budget.filters import ItemViewFilter
from budget.filters import PackageViewFilter
from budget.filters import PrintPublicationViewFilter
from budget.models import ContentPlacement
from budget.models import Headline
from budget.models import HeadlineVote
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
from budget.paginators import ItemViewPagination
from budget.serializers import ContentPlacementSerializer
from budget.serializers import HeadlineSerializer
from budget.serializers import HeadlineVoteSerializer
from budget.serializers import ItemSerializer
from budget.serializers import PackageSerializer
from budget.serializers import PrintPublicationSerializer


class MainBudgetView(TemplateView):
    """TK."""
    template_name = 'budget/index.html'

    def get_context_data(self, *args, **kwargs):
        context = super(MainBudgetView, self).get_context_data(*args, **kwargs)

        hostURL = self.request.META['HTTP_HOST']
        aliased_origins = getattr(settings, 'BUDGET_ALIASED_ORIGINS', [])

        context.update({
            'budget_name': getattr(settings, 'BUDGET_TOOL_NAME', 'Budget'),
        })

        # Config AJAX URL.
        config_url = reverse('budget:config')
        if hostURL is not None:
            formatted_url = '{}://{}'.format(self.request.scheme, hostURL)

            if formatted_url in aliased_origins:
                root_budget_url = reverse(
                    'budget:main-budget-view',
                    kwargs={'path': ''}
                )
                config_url = config_url.lstrip(root_budget_url)

                if config_url[0] != '/':
                    config_url = '/{}'.format(config_url)

        context.update({
            'config_url': config_url,
        })

        return context


class ExceptionView(View):
    def get(self, request, *args, **kwargs):
        return 1 / 0


class ConfigView(View):
    def get(self, request, *args, **kwargs):
        originURL = request.GET.get('origin')
        aliased_origins = getattr(settings, 'BUDGET_ALIASED_ORIGINS', [])

        empty_admin_case = [['', 'test@example.com']]
        django_admins = getattr(settings, 'ADMINS', empty_admin_case)
        default_admin_email = django_admins[0][1]

        actual_admin_email = getattr(
            settings,
            'BUDGET_ADMIN_EMAIL',
            default_admin_email
        )

        api_bases = {
            'auth': reverse('apiauth:api-root'),
            'budget': reverse('budget:api-root'),
            'staff': reverse('editorial_staff:api:root'),
        }

        explicit_api_bases = getattr(settings, 'BUDGET_API_CONFIGS', {})

        for key, val in explicit_api_bases.items():
            if callable(val):
                raw_value = val(reverse)
            else:
                raw_value = val

            api_bases[key] = raw_value

        for k, endpoint_url in api_bases.items():
            if originURL is not None and originURL in aliased_origins:
                endpoint_url_parts = urlparse(endpoint_url)

                aliased_url = getattr(settings, 'BUDGET_ALIASED_API_URL', '')
                aliased_url_parts = urlparse(aliased_url)

                final_url = urlunparse(endpoint_url_parts._replace(
                    scheme=aliased_url_parts.scheme,
                    netloc=aliased_url_parts.netloc,
                    path=endpoint_url_parts.path.replace('//', '/')
                ))

                api_bases[k] = final_url

        if originURL is not None and originURL in aliased_origins:
            root_url = None
        else:
            root_url = reverse('budget:main-budget-view', kwargs={'path': ''})

        return JsonResponse({
            'adminEmail': actual_admin_email,
            'apiBases': api_bases,
            'branding': {
                'mastheadLogoAltText': getattr(
                    settings,
                    'BUDGET_ORGANIZATION_NAME',
                    'Your organization here'
                ),
                'mastheadLogoURL': static(getattr(
                    settings,
                    'BUDGET_ORGANIZATION_LOGO_PATH',
                    'budget/images/your_name_here.svg'
                )),
                'siteLogoAltText': getattr(
                    settings,
                    'BUDGET_TOOL_NAME',
                    'Budget'
                ),
                'siteLogoURL': static(getattr(
                    settings,
                    'BUDGET_TOOL_LOGO_PATH',
                    'budget/images/budget_logo.svg'
                )),
                'printLogoURL': static(getattr(
                    settings,
                    'BUDGET_PRINT_LOGO_PATH',
                    'budget/images/the-daily.svg'
                )),
            },
            'defaultTimezone': timezone.get_default_timezone_name(),
            'externalURLs': json.dumps(
                getattr(settings, 'BUDGET_EXTERNAL_URLS', {})
            ),
            'imageAssets': {
                'notepad': static('budget/images/notepad.svg'),
                'slackProgress': static('budget/images/slack-in-progress.png'),
            },
            'maxItemsToLoad': getattr(settings, 'BUDGET_API_MAX_ITEMS', 500),
            'printSlugName': getattr(settings, 'BUDGET_PRINT_SLUG_NAME', None),
            'showHeadlines': getattr(settings, 'BUDGET_SHOW_HEADLINES', False),
            'rootURL': root_url
        })


##########
# Mixins #
##########

class CsrfExemptSessionAuthentication(SessionAuthentication):
    """A custom auth subclass to disable CSRF on the session auth middleware"""
    def enforce_csrf(self, request):
        return


class SessionAndTokenAuthedViewSet(object):
    """Mixin to add both token and Django session auth to a REST viewset"""
    authentication_classes = (CsrfExemptSessionAuthentication,
                              TokenAuthentication,)
    permission_classes = (IsAuthenticated,)


class CamelCasedViewSet(object):
    renderer_classes = (BrowsableAPIRenderer, CamelCaseJSONRenderer,)
    parser_classes = (CamelCaseJSONParser,)


############
# Viewsets #
############


class PackageViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                     viewsets.ModelViewSet):
    serializer_class = PackageSerializer
    filter_class = PackageViewFilter

    queryset = Package.objects.distinct().prefetch_related(
        'additional_content',
        'headline_candidates',
        # 'print_section',
    )

    def get_object(self):
        """Attempt to retrieve the package by multiple means.

        If the PK passed is numeric, try getting the package by its ID.
        Otherwise, use the slug instead.
        """
        queryset = self.get_queryset()
        queryset = self.filter_queryset(queryset)

        try:
            int(self.kwargs.get('pk', ''))
            return get_object_or_404(queryset, **self.kwargs)
        except ValueError:
            try:
                return queryset.get_by_slug(self.kwargs['raw_slug'])
            except Package.DoesNotExist:
                raise Http404


class ItemViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                  viewsets.ModelViewSet):
    queryset = Item.objects.select_related(
        'primary_for_package',
        'additional_for_package'
    )
    serializer_class = ItemSerializer
    filter_class = ItemViewFilter
    pagination_class = ItemViewPagination


class ContentPlacementViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                              viewsets.ModelViewSet):
    serializer_class = ContentPlacementSerializer
    filter_class = ContentPlacementFilter

    queryset = ContentPlacement.objects.distinct()


class HeadlineViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                      viewsets.ModelViewSet):
    queryset = Headline.objects.all().order_by('-pk')
    serializer_class = HeadlineSerializer
    filter_class = HeadlineViewFilter


class HeadlineVoteViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                          viewsets.ModelViewSet):
    queryset = HeadlineVote.objects.all().order_by('-pk')
    serializer_class = HeadlineVoteSerializer


class PrintPublicationViewSet(SessionAndTokenAuthedViewSet, CamelCasedViewSet,
                              viewsets.ReadOnlyModelViewSet):
    queryset = PrintPublication.objects.all().prefetch_related('sections')
    serializer_class = PrintPublicationSerializer
    filter_class = PrintPublicationViewFilter



# Imports from django.  # NOQA
from django.db.models.signals import pre_save
from django.dispatch import receiver


# Imports from other dependencies.
from editorial_staff.models import Hub


# Imports from budget.
from budget.models import Package


@receiver(pre_save, dispatch_uid='budget_package_save', sender=Package)
def update_package_vertical_on_save(sender, instance, **kwargs):
    """Update package's vertical before saving."""
    instance.vertical = Hub.objects.get(slug=instance.hub).vertical.slug



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-27 15:06
from __future__ import unicode_literals

import budget.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0012_merge'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='package',
            options={},
        ),
        migrations.AlterField(
            model_name='package',
            name='hub',
            field=models.CharField(max_length=100, validators=[budget.validators.hub_exists]),
        ),
    ]
